@@ -9,6 +9,7 @@ except ImportError:
 
 import requests
 from requests.exceptions import HTTPError, ConnectionError, Timeout
+from requests.adapters import HTTPAdapter
 
 try:
     # noinspection PyUnresolvedReferences
@@ -38,12 +39,15 @@ SESSION_TIME_TO_LIVE = 600  # In seconds. None - live forever, 0 - one-time
 RETRY_ON_ERROR = False
 RETRY_TIMEOUT = 2
 MAX_RETRIES = 15
+RETRY_ENGINE = 1
 
 CUSTOM_SERIALIZER = None
+CUSTOM_REQUEST_SENDER = None
 
 ENABLE_MIDDLEWARE = False
 
 
+
 def _get_req_session(reset=False):
     if SESSION_TIME_TO_LIVE:
         # If session TTL is set - check time passed
@@ -105,36 +109,44 @@ def _make_request(token, method_name, method='get', params=None, files=None):
     params = params or None # Set params to None if empty
 
     result = None
-    if RETRY_ON_ERROR:
+    if RETRY_ON_ERROR and RETRY_ENGINE == 1:
         got_result = False
         current_try = 0
-        
         while not got_result and current_try<MAX_RETRIES-1:
             current_try+=1
-            
             try:
                 result = _get_req_session().request(
                     method, request_url, params=params, files=files,
                     timeout=(connect_timeout, read_timeout), proxies=proxy)
                 got_result = True
-                
             except HTTPError:
                 logger.debug("HTTP Error on {0} method (Try #{1})".format(method_name, current_try))
                 time.sleep(RETRY_TIMEOUT)
-                
             except ConnectionError:
                 logger.debug("Connection Error on {0} method (Try #{1})".format(method_name, current_try))
                 time.sleep(RETRY_TIMEOUT)
-                
             except Timeout:
                 logger.debug("Timeout Error on {0} method (Try #{1})".format(method_name, current_try))
                 time.sleep(RETRY_TIMEOUT)
-            
-        
         if not got_result:
             result = _get_req_session().request(
                     method, request_url, params=params, files=files,
                     timeout=(connect_timeout, read_timeout), proxies=proxy)
+    elif RETRY_ON_ERROR and RETRY_ENGINE == 2:
+        http = _get_req_session()
+        retry_strategy = requests.packages.urllib3.util.retry.Retry(
+            total=MAX_RETRIES,
+        )
+        adapter = HTTPAdapter(max_retries=retry_strategy)
+        for prefix in ('http://', 'https://'):
+            http.mount(prefix, adapter)
+        result = http.request(
+            method, request_url, params=params, files=files,
+            timeout=(connect_timeout, read_timeout), proxies=proxy)
+    elif CUSTOM_REQUEST_SENDER:
+        result = CUSTOM_REQUEST_SENDER(
+            method, request_url, params=params, files=files,
+            timeout=(connect_timeout, read_timeout), proxies=proxy)
     else:
         result = _get_req_session().request(
             method, request_url, params=params, files=files,
@@ -966,7 +978,7 @@ def set_chat_permissions(token, chat_id, permissions):
     return _make_request(token, method_url, params=payload, method='post')
 
 
-def create_chat_invite_link(token, chat_id, expire_date, member_limit):
+def create_chat_invite_link(token, chat_id, name, expire_date, member_limit, creates_join_request):
     method_url = 'createChatInviteLink'
     payload = {
         'chat_id': chat_id
@@ -979,11 +991,15 @@ def create_chat_invite_link(token, chat_id, expire_date, member_limit):
             payload['expire_date'] = expire_date
     if member_limit:
         payload['member_limit'] = member_limit
+    if creates_join_request is not None:
+        payload['creates_join_request'] = creates_join_request
+    if name:
+        payload['name'] = name
 
     return _make_request(token, method_url, params=payload, method='post')
 
 
-def edit_chat_invite_link(token, chat_id, invite_link, expire_date, member_limit):
+def edit_chat_invite_link(token, chat_id, invite_link, name, expire_date, member_limit, creates_join_request):
     method_url = 'editChatInviteLink'
     payload = {
         'chat_id': chat_id,
@@ -998,6 +1014,10 @@ def edit_chat_invite_link(token, chat_id, invite_link, expire_date, member_limit
 
     if member_limit is not None:
         payload['member_limit'] = member_limit
+    if name:
+        payload['name'] = name
+    if creates_join_request is not None:
+        payload['creates_join_request'] = creates_join_request
 
     return _make_request(token, method_url, params=payload, method='post')
 
@@ -1017,6 +1037,24 @@ def export_chat_invite_link(token, chat_id):
     return _make_request(token, method_url, params=payload, method='post')
 
 
+def approve_chat_join_request(token, chat_id, user_id):
+    method_url = 'approveChatJoinRequest'
+    payload = {
+        'chat_id': chat_id,
+        'user_id': user_id
+    }
+    return _make_request(token, method_url, params=payload, method='post')
+
+
+def decline_chat_join_request(token, chat_id, user_id):
+    method_url = 'declineChatJoinRequest'
+    payload = {
+        'chat_id': chat_id,
+        'user_id': user_id
+    }
+    return _make_request(token, method_url, params=payload, method='post')
+
+
 def set_chat_photo(token, chat_id, photo):
     method_url = 'setChatPhoto'
     payload = {'chat_id': chat_id}
@@ -1661,4 +1699,5 @@ class ApiTelegramException(ApiException):
             result)
         self.result_json = result_json
         self.error_code = result_json['error_code']
+        self.description = result_json['description']
         
